/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
/* jshint esnext: true */
/* jshint multistr: true */
/**
    Copyright (C) 2017 Jason Gray <jasonlevigray3@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

const Gio = imports.gi.Gio;
const Lang = imports.lang;
const GnomeSession = imports.misc.gnomeSession;

const DBusProxy = Gio.DBusProxy.makeProxyWrapper('<node> \
<interface name="org.freedesktop.DBus"> \
  <method name="ListNames">\
    <arg type="as" direction="out" />\
  </method>\
  <signal name="NameOwnerChanged"> \
    <arg type="s" direction="out" name="name" /> \
    <arg type="s" direction="out" name="oldOwner" /> \
    <arg type="s" direction="out" name="newOwner" /> \
  </signal> \
</interface> \
</node>');

const MprisPlayerProxy = Gio.DBusProxy.makeProxyWrapper('<node> \
<interface name="org.mpris.MediaPlayer2.Player"> \
  <method name="Pause" /> \
  <method name="Stop" /> \
  <method name="Play" /> \
  <property name="CanPause" type="b" access="read" />\
  <property name="CanPlay" type="b" access="read" />\
  <property name="PlaybackStatus" type="s" access="read" />\
</interface> \
</node>');

const SESSION_BUS = Gio.DBus.session;
const MPRIS_PREFIX = 'org.mpris';
const MPRIS_PLAYER_2_IFACE = '/org/mpris/MediaPlayer2';
const PLAYING = 'Playing';
const PAUSED = 'Paused';


const Morphine = new Lang.Class({
    Name: 'Morphine',

    _init: function() {
      this._players = [];
      this._dbus = null;
      this._dbusNameOwnerChangedId = null;
      this._presence = null;
      this._statusChangedId = null;
      new DBusProxy(SESSION_BUS,
                    'org.freedesktop.DBus',
                    '/org/freedesktop/DBus',
                    Lang.bind(this, this._onDBusProxyReady));
    },

    _pausePlayers: function() {
      // I would prefer to just create the proxy once and use it
      // for the life of the player, but players have a nasty
      // habit of not updating properties when they change.
      // This way we get up to date properties every time.
      this._players.forEach(Lang.bind(this, function(player) {
        new MprisPlayerProxy(SESSION_BUS,
                             player._name,
                             MPRIS_PLAYER_2_IFACE,
                             Lang.bind(this, function(proxy) {
                               player._paused = false;
                               try {
                                 if (proxy.PlaybackStatus == PLAYING) {       
                                   if (proxy.CanPause) {
                                     player._paused = true;
                                     proxy.PauseRemote();
                                   }
                                   else {
                                     proxy.StopRemote();
                                   }
                                 }
                               }
                               catch (error) {
                                 this._logError('_pausePlayers', player._name, error);
                               }
                             }));
      }));
    },

    _resumePlayers: function() {
      // Resume playback if the session idle fade out is interrupted and we paused the player.
      // Don't resume if the player was stopped. Stopped is an inconsistent state.
      // It's unclear what will happen if we hit play. Will the playlist start over?
      // Will the song start over? Will it do anything at all? 
      this._players.forEach(Lang.bind(this, function(player) {
        if (player._paused) {
          player._paused = false;
          new MprisPlayerProxy(SESSION_BUS,
                               player._name,
                               MPRIS_PLAYER_2_IFACE,
                               Lang.bind(this, function(proxy) {
                                 try {
                                   if (proxy.CanPlay && proxy.PlaybackStatus == PAUSED) {       
                                     proxy.PlayRemote();
                                   }
                                 }
                                 catch (error) {
                                   this._logError('_resumePlayers', player._name, error);
                                 }
                               }));
        }
      }));
    },

    _onNameOwnerChanged: function(proxy, sender, [name, oldOwner, newOwner]) {
      if (!name.startsWith(MPRIS_PREFIX)) {
        return;
      }
      let index = this._players.findIndex(function(player) {
        return player._name == name;
      });
      if (newOwner && !oldOwner && index == -1) {
        this._players.push({_name: name});
      }
      if (oldOwner && !newOwner && index != -1) {
        this._players.splice(index, 1);
      }
    },

    _onDBusProxyReady: function(proxy) {
      this._dbus = proxy;
      this._dbus.ListNamesRemote(Lang.bind(this, function(names) {
        names[0].forEach(Lang.bind(this, function(name) {
          if (name.startsWith(MPRIS_PREFIX)) {
            this._players.push({_name: name});
          }
        }));
      }));

      this._dbusNameOwnerChangedId = this._dbus.connectSignal('NameOwnerChanged',
                                       Lang.bind(this, this._onNameOwnerChanged));

      this._presence = new GnomeSession.Presence();
      this._statusChangedId = this._presence.connectSignal('StatusChanged',
                                Lang.bind(this, function(proxy, senderName, [status]) {
                                  if (status == GnomeSession.PresenceStatus.IDLE) {
                                    this._pausePlayers();
                                  }
                                  else if (status == GnomeSession.PresenceStatus.AVAILABLE) {
                                    this._resumePlayers();
                                  } 
                                })); 
    },

    _logError: function(func, busName, error) {
      global.log('function:', func, 'Player bus name:',
        busName, 'Error:', error);
    },

    destroy: function() {
      if (this._dbusNameOwnerChangedId) {
        this._dbus.disconnectSignal(this._dbusNameOwnerChangedId);
      }
      if (this._statusChangedId) {
        this._presence.disconnectSignal(this._statusChangedId);
      }
      this._players = null;
      this._dbus = null;
      this._dbusNameOwnerChangedId = null;
      this._presence = null;
      this._statusChangedId = null;
    }
});

let _extension = null;

//We don't need to init anything...
function init() {}

function enable() {
    _extension = new Morphine();
}

function disable() {
    if (_extension) {
      _extension.destroy();
      _extension = null;
    }
}
